#!/usr/bin/perl
#
# $Id: netdog.pl,v 1.2 2003/04/01 11:24:26 magnus Exp $
# (C) Magnus Bodin; magnus@bodin.org; http://x42.com/
# License: GPL 2.0
#
# "Manual" at the bottom in this file.
# Either read POD or do: perldoc netdog.pl
#

use strict;

#
# Set this to whatever you don't want to pass on the commandline
#
my $DEFAULTCONFFILE = '/etc/netdog.conf';
my $CONFFILE = shift(@ARGV) || $DEFAULTCONFFILE;

my $opt = readconf($CONFFILE);
$$opt{bindir} =~ s#/$##;
$$opt{bindir} ||= '/usr/local/bin';

$$opt{tcpping} ||= 'tcp_ping.pl';
$$opt{icmpping} ||= 'icmp_ping.pl';


sub readconf ($)
{
    my ($file) = @_;
    my $opt = {};
    my($string, @option, $state, %optindex);

    $state = 1;
    $$opt{logfile} = '';

    open(CONF, $file) or die "Couldn't open configfile '$file': $!";
    for my $line (<CONF>)
    {
	# remove comments and strip leading+trailing spaces
	$line =~ s/#.+$//g; $line =~ s/\s+$//; $line =~ s/^\s+//;
	
	if ($line =~ /^\s*(\w+)\s*(.+)\s*$/)
	{
	    my ($name, $value) = (lc $1, $2);

        if ($name =~ m#^(logfile|bindir|tcpping|icmpping)$#i)
	    {
		    $$opt{$1} = $value;
        }
	    elsif ($name =~ m#^(host|on)$#i)
	    {
	        push @{$$opt{$state}->{$1}}, $value;
	    }
	    elsif ($name =~ m#^(frequency)$#i)
	    {
		    $$opt{$state}->{$1} = $value;
	    }
        elsif ($name eq 'state')
	    {
	        $state = $value;
	    }
	}
    }

    return $opt;
}

sub do_ping ($$)
{
	my ($URI, $opt) = @_;
	my @result;

	if ($URI =~ m#^icmp://(.+)/?$#i)
	{
		my $host = $1;
        my $p = "$$opt{bindir}/$$opt{icmpping}";
        die "Couldn't find $p: $!" unless -x $p;
		@result = split /\n/, `$$opt{bindir}/$$opt{icmpping} $host`;
	}
	else
	{
        my $p = "$$opt{bindir}/$$opt{tcpping}";
        die "Couldn't find $p: $!" unless -x $p;
		@result = split /\n/, `$$opt{bindir}/$$opt{tcpping} $URI`;
	}

	my $pl = $result[0];
	$pl = 100 if $pl < 0;

	return [ $pl, @result ];
}


my $state;
my $nextstate = 1;
my $prevstate = 1;

my @plhistory = ( 0 );

while (1)
{
	$prevstate = $state;
	$state = $nextstate; 

	my $starttime = time();
	my $dolog = 0;

	# loop through all hosts and determine total connectivity
	my ($hostc, $totpl);
	my @loglines = ();
	for my $host (@{$$opt{$state}->{host}})
	{
		my $ping = do_ping($host, $opt);
		$$ping[5] ||= 0;
		push @loglines, time()." ping $host $$ping[0] $$ping[5]";
		$totpl += $$ping[0];
		$hostc++;
	}

	my $currpl = 100;
	if ($hostc)
	{
		$currpl = $totpl / $hostc;
	}

	# put currpl into the plhistory vector
	unshift @plhistory, $currpl;
	pop(@plhistory) if @plhistory > 10;

	$nextstate = 1 if $currpl < 100;

	# walk through all action items and check if actionstate > $lastpl
	my @actions = @{$$opt{$state}->{on}};
	for my $action (@actions)
	{
		my $doit = 0;
		if ($action =~ /^(\d\S*)\s+(.+)$/)
		{
			my $plevel = $1; 
			$action = $2;
			$doit = 1;
			my $plevelf = 1;
			($plevel,$plevelf) = ($1, $2)
				if $plevel =~ m#^(\d+)/(\d+)$#;

			# now check the action against plevel,f,history
			for my $i (0..($plevelf-1))
			{
  	 			if ($plhistory[$i] < $plevel)
				{
					$doit = 0;
				}
			}
			$doit = 1 if $plevel = 0;
		}
		elsif ($action =~ /^([\-\+])\s+(.+)$/)
		{
			my $flank = $1;
			$action = $2;
			$doit = 0;
			$doit = 1 if ($prevstate < $state) && ($flank eq '+');
			$doit = 1 if ($prevstate > $state) && ($flank eq '-');
		}

		if ($doit) 
		{
			if ($action =~ /^state\s+(\d+)$/i)
			{
				$nextstate = $1;
				unshift @plhistory,0;
			}
			elsif ($action =~ /^log(.*)$/i)
			{
				$dolog = 1;
			}
			elsif ($action =~ /^exec\s+(.+)$/)
			{
				my $result = `$1`;
				push @loglines,
				   time()." exec $1";
			}
		}

	}

	if ($nextstate ne $state)
	{
		push @loglines, time()." state $state -> $nextstate";
	}
	
	if ($dolog)
	{
		open LOGFILE, ">>$$opt{logfile}" or die "Couldn't open logfile: $!";
		print LOGFILE join("\n",@loglines),"\n";
		close LOGFILE;
	}

	my $sleeptime = $starttime + $$opt{$state}->{frequency} - time();
	if (($nextstate == $state) && ($sleeptime > 0))
	{
		sleep $sleeptime;
	}

}

__END__

=head1 NAME

netdog.pl - network watchdog

=head1 SYNOPSIS

 nohup ./netdog.pl &

=head1 DESCRIPTION

I<netdog.pl> is a watchdog script to be used e.g. to monitor if a host is
reachable or not. It is basically a statemachine that has necessary actions
at each state which is to "ping" hosts around the net. Then conditional actions
can be triggered by a certain level of total network-downness (packetloss).
These actions can be state changes, logging and executing external programs.

If this program is run by root, then it is important that ALL files used
(both configuration files and external run programs) and their directories
are controlled by root. Otherwise you will put yourself and your users in
an unsecure situation.

=head1 EXAMPLES

This is the only documention right now regarding the configuration file.

  # comments are lines that start with the character '#'
  #
  # This is netdog.conf
  # It is by default searched for in /etc
  # If you put it elsewhere, pass it as the only
  # command line argument to netdog.pl like this: 
  #
  #    netdog.pl /usr/local/etc/netdog.conf
  #

  # Define in which directory we will find the external programs
  bindir /usr/local/share/netdog/bin/

  # logfile name 
  logfile /var/log/netdog.log 


  # the rest of the configuration file is divided in the different
  # "states" in which netdog can be in.
  # 
  # netdog always starts in state 1



  #
  ## DEFINITION OF STATE 1
  # 
  state 1
 
  # When in state 1, repeat the "ping-loop" every
  # five minutes (300 seconds)
  frequency 300

  # ping google with "normal" icmp-ping
  host icmp://www.google.com

  # http-connect-ping to amazon
  host http://www.amazon.com

  # tcp-connect to your external ssh-box 'mybox' (in /etc/hosts)
  host tcp://mybox

  # tcp-connect to your ISP:s nameserver
  host tcp://10.0.0.2

  # if packetloss is >=0 (i.e. always) log the
  # loglines to the logfile
  on 0 log

  # if 90% packedloss execute the external script ninety.sh
  on 90 exec ninety.sh
 
  # if 100% packetloss on two subsequent rounds execute external
  # script restartinterface.sh and also go to state 2
  on 100/2 exec restartinterface.sh
  on 100/2 state 2

  # if we reached state 1 by DECREMENTING the state
  # (i.e. 2->1 or 3->1) then execute the "WE ARE GOOD AGAIN"
  # -script to celebrate... (Maybe ip has changed and we want
  # to notify someone about this)
  on - exec checkipandfixdyndns.sh

  #
  ## END OF STATE 1
  #




  #
  ## DEFINITION OF STATE 2
  # 
  state 2

  # When in state, repeat everything every 30 seconds (we want to go back
  # to state 1 as soon as possible)
  frequency 30

  # Just check two smtp-servers 
  host smtp://fluff
  host smtp://getz

  # Always log
  on 0 log

  # if 100% on two subsequent rounds execute 'restart ADSL-modem.sh'
  on 100/2 exec restart_ADSL_modem.sh
 
  # if it is still down after next round (three in a row) go to state 3
  on 100/3 state 3

  # If we just came to state 2 from state 1 (i.e. state was incremented)
  # then execute restart_network_interface.sh
  on + exec restart_network_interface.sh

  #
  ## END OF STATE 2
  #


  #
  ## DEFINITION OF STATE 3
  #
  state 3

  # Keep a fast pace
  frequency 15 
  host tcp://10.0.0.2
  on + exec send_sms_to_administratot.sh
  on 0 log


=head1 BUGS

No known bugs

=head1 THANKS

Lundaspam, Janek Hellqvist, Telia

=head1 COPYRIGHT

Copyright 1999-2003 Magnus Bodin

This library is free software; you can redistribute it and/or
modify it under GPL 2.0

=cut
